@@ -159,6 +159,15 @@ public class BookKeeper implements OpenCallback, CreateCallback {
     return zk;
   }
 
+  /**
+   * Get the BookieClient, currently used for doing bookie recovery.
+   * 
+   * @return BookieClient for the BookKeeper instance.
+   */
+  public BookieClient getBookieClient() {
+      return bookieClient;
+  }
+  
   /**
    * Creates a new ledger asynchronously. To create a ledger, we need to specify
    * the ensemble size, the quorum size, the digest type, a password, a callback
